@@ -23,9 +23,13 @@ channel_genes = dataset.create_channel_gene_table()
 
 """
 import json
+import multiprocessing as mp
+import pickle as pkl
+import warnings
 from dataclasses import dataclass
+from functools import partial
 from pathlib import Path
-from typing import Dict
+from typing import Dict, Optional
 
 import numpy as np
 import pandas as pd
@@ -81,7 +85,19 @@ class SpotDetection:
 
     channel: str
     spots_file: Path
-    stats_files: dict
+    stats_files: Dict[str, Path]
+
+
+@dataclass
+class TileAlignmentFiles:
+    """Dataclass to hold paths to tile alignment XML files."""
+
+    raw_single_xml: Optional[Path] = None
+    raw_single_tile_subset_xml: Optional[Path] = None
+    raw_spot_xml: Optional[Path] = None
+    raw_spot_tile_subset_xml: Optional[Path] = None
+    pc_xml: Optional[Path] = None
+    ip_xml: Optional[Path] = None
 
 
 @dataclass
@@ -94,6 +110,7 @@ class SegmentationFiles:
         str, Path
     ]  # Dictionary mapping resolution keys to segmentation mask paths
     cell_centroids: Path = None  # Path to cell centroids file
+    metrics_path: Path = None
 
     def __post_init__(self):
         """Initialize empty dictionary if not provided."""
@@ -118,43 +135,59 @@ class HCRRound:
     def __init__(
         self,
         round_key: str,
-        spot_files: SpotFiles,
-        zarr_files: ZarrDataFiles,
+        name: str,
+        spot_files: "SpotFiles",
+        zarr_files: "ZarrDataFiles",
+        segmentation_files: "SegmentationFiles",
+        spot_detection_files: Dict[str, "SpotDetection"],
         processing_manifest: dict,
-        name: str = None,
-        segmentation_files: SegmentationFiles = None,
-        spot_detection_files: Dict[str, SpotDetection] = None,
+        tile_alignment_files: "TileAlignmentFiles" = None,
     ):
         """
-        Initialize an HCRRound object.
+        Initialize HCRRound.
 
         Parameters:
         -----------
         round_key : str
             The identifier for this round (e.g., 'R1', 'R2')
+        name : str
+            Name of the round. Folder name where data is stored.
         spot_files : SpotFiles
             Spot files for this round
         zarr_files : ZarrDataFiles
             Zarr files for this round
-        processing_manifest : dict
-            Processing manifest data for this round
-        name : str, optional
-            Dataset name for this round
         segmentation_files : SegmentationFiles, optional
             Segmentation files for this round
         spot_detection_files : Dict[str, SpotDetection], optional
             Spot detection files for this round, mapping channel to SpotDetection
+        processing_manifest : dict
+            Processing manifest data for this round
+        tile_alignment_files : TileAlignmentFiles, optional
+            Tile alignment files for this round
         """
         self.round_key = round_key
         self.name = name
         self.spot_files = spot_files
         self.zarr_files = zarr_files
-        self.processing_manifest = processing_manifest
         self.segmentation_files = segmentation_files
-        self.spot_detection_files = spot_detection_files or {}
+        self.spot_detection_files = spot_detection_files
+        self.processing_manifest = processing_manifest
+        self.tile_alignment_files = tile_alignment_files
 
-    def get_channels(self):
-        """Get list of available channels for this round."""
+    def get_channels(self, data_type="fused"):
+        """
+        Get list of available channels for this round.
+
+        Parameters:
+        -----------
+        data_type : str
+            Type of data ('fused', 'corrected', 'raw')
+
+        Returns:
+        --------
+        list
+            List of available channels for this round
+        """
         return self.zarr_files.get_channels()
 
     def has_channel(self, channel):
@@ -237,7 +270,20 @@ class HCRRound:
             )
 
         mask_path = self.segmentation_files.segmentation_masks[resolution_key]
-        return zarr.open(mask_path, mode="r")["0"]
+
+        if resolution_key == "0":
+            # Load the standard segmentation mask
+            return zarr.open(mask_path, mode="r")["0"]
+        elif resolution_key == "2":
+            # sometimes the original resolution mask is stored in a group....
+            zarr_file = zarr.open(mask_path, mode="r")
+            if isinstance(zarr_file, zarr.Array):
+                return zarr_file
+            elif isinstance(zarr_file, zarr.Group):
+                if "0" in zarr_file:
+                    return zarr_file["0"]
+                else:
+                    raise ValueError(f"No '0' array found in {mask_path}")
 
     def load_cell_centroids(self):
         """
@@ -259,9 +305,21 @@ class HCRRound:
 
         return np.load(self.segmentation_files.cell_centroids)
 
-    def get_cell_info(self):
+    def get_cell_info(self, source="mixed_cxg"):
         """
-        Get cell information for this round.
+        Get cell information.
+
+        segmentation:
+            contains every mask from segmentation,
+            columns: cell_id, x_centroid, y_centroid, z_centroid
+        mixed_cxg:
+            contains subset of masks, that had a detected spot
+            columns: cell_id, volume, x_centroid, y_centroid, z_centroid
+
+        Parameters:
+        -----------
+        source : str
+            Source of cell information ('mixed_cxg' or 'unmixed_cxg')
 
         Returns:
         --------
@@ -269,14 +327,70 @@ class HCRRound:
             DataFrame containing cell_id, volume, and centroid coordinates
         """
         # Read data for this round
-        df = pd.read_csv(self.spot_files.unmixed_cxg)
+        if source not in ["mixed_cxg", "unmixed_cxg", "segmentation"]:
+            raise ValueError("Source must be 'mixed_cxg', 'unmixed_cxg', or 'segmentation'")
 
-        # Keep only the columns we want
-        cols_to_keep = ["cell_id", "volume", "x_centroid", "y_centroid", "z_centroid"]
-        df_cells = df[cols_to_keep].drop_duplicates()
+        if source == "unmixed_cxg":
+            print(f"Loading unmixed cxg for round {self.round_key}")
+            try:
+                df = pd.read_csv(self.spot_files.unmixed_cxg)
+            except Exception as e:
+                print(f"Warning: Error reading unmixed cxg file: {e}")
+                return pd.DataFrame()  # Return empty DataFrame if file does not exist
 
+            # add warning, getting cell info from unmixed cxg
+            warnings.warn(
+                "Getting cell info from unmixed cxg file. Does not include all segmentation masks."
+            )
+
+            # Keep only the columns we want
+            cols_to_keep = ["cell_id", "volume", "x_centroid", "y_centroid", "z_centroid"]
+            df_cells = df[cols_to_keep].drop_duplicates()
+        elif source == "mixed_cxg":
+            print(f"Loading mixed cxg for round {self.round_key}")
+            try:
+                df = pd.read_csv(self.spot_files.mixed_cxg)
+            except Exception as e:
+                print(f"Warning: Error reading mixed cxg file: {e}")
+                return pd.DataFrame()  # Return empty DataFrame if file does not exist
+
+            # Keep only the columns we want
+            cols_to_keep = ["cell_id", "volume", "x_centroid", "y_centroid", "z_centroid"]
+            df_cells = df[cols_to_keep].drop_duplicates()
+        elif source == "segmentation":
+            # load centroids from segmentation files
+            centroids = self.load_cell_centroids()
+            df_cells = pd.DataFrame(
+                centroids, columns=["z_centroid", "y_centroid", "x_centroid", "cell_id"]
+            )
+            # Retain the original cell_id values from the centroids data
+        print(f"Number of cells in {source} for round {self.round_key}: {len(df_cells)}")
         return df_cells
 
+    def load_spots(self, table_type="mixed_spots"):
+        """
+        Load spots for this round (non-multiprocessing version).
+
+        Parameters:
+        -----------
+        table_type : str
+            Type of spots to load ('mixed_spots' or 'unmixed_spots')
+
+        Returns:
+        --------
+        pd.DataFrame
+            DataFrame with spots data including round column
+        """
+        print(f"Loading {table_type} for round {self.round_key}")
+
+        spot_file_path = getattr(self.spot_files, table_type)
+
+        with open(spot_file_path, "rb") as f:
+            spots_data = pkl.load(f)
+            spots_data["round"] = self.round_key
+
+        return spots_data
+
     def __dir__(self):
         """
         Return a list of valid attributes and methods for this HCRRound.
@@ -396,10 +510,12 @@ class HCRDataset:
         """Check if dataset contains a specific round."""
         return round_key in self.rounds
 
-    def get_cell_info(self, round_key="R1"):
+    def get_cell_info(self, source="mixed_cxg"):
         """
         Get cell information from a specific round.
 
+        Usually just get R1 cell ids.
+
         Parameters:
         -----------
         round_key : str
@@ -410,10 +526,72 @@ class HCRDataset:
         pd.DataFrame
             DataFrame containing cell_id, volume, and centroid coordinates
         """
-        if round_key not in self.rounds:
-            raise ValueError(f"Round {round_key} not found")
 
-        return self.rounds[round_key].get_cell_info()
+        if source == "segmentation":
+            # only R1 has segmentation centroids
+            return self.rounds["R1"].get_cell_info(source=source)
+
+        if source == "mixed_cxg" or source == "unmixed_cxg":
+            # Concatenate cell info from all rounds
+            all_cells = []
+            for r_key, round_obj in self.rounds.items():
+                cells = round_obj.get_cell_info(source=source)
+                # cells['round'] = r_key
+                all_cells.append(cells)
+
+                # get the unique cell ids
+            all_cells_df = pd.concat(all_cells, ignore_index=True)
+            all_cells_df = all_cells_df.drop_duplicates(subset=["cell_id"]).reset_index(drop=True)
+
+            return all_cells_df
+
+    def load_all_rounds_spots_mp(self, table_type="mixed_spots"):
+        """
+        Load all spots from the dataset in parallel.
+
+        Parameters:
+        -----------
+        table_type : str
+            Type of spots to load ('mixed_spots' or 'unmixed_spots')
+
+        Returns:
+        --------
+        pd.DataFrame
+            DataFrame containing all spots from all rounds
+        """
+        # Create list of (round_key, round_obj) tuples for multiprocessing
+        round_items = list(self.rounds.items())
+
+        # Use multiprocessing to load spots from all rounds
+        pool = mp.Pool(processes=min(len(self.rounds), mp.cpu_count()))
+        process_round = partial(_load_spots_for_round, table_type=table_type)
+        all_spots_list = pool.map(process_round, round_items)
+        pool.close()
+        pool.join()
+
+        # Concatenate all DataFrames
+        all_spots = pd.concat(all_spots_list, ignore_index=True)
+        print(f"Number of {table_type}: {len(all_spots):.3e}")
+
+        return all_spots
+
+    def load_all_rounds_spots_coreg(self, ophys_mfish_match_df, table_type="mixed_spots"):
+        """Wrapper function around load_all_rounds_spots_mp to filter for coregistered spots.
+        Parameters:
+        - dataset: The HCR dataset containing the rounds and mixed spots.
+        - ophys_mfish_match_df: DataFrame containing coregistered spots with 'ls_id' column.
+        - table_type: Type of spots to load, default is 'mixed_spots'.
+        Returns:
+        - DataFrame containing coregistered mixed spots.
+        """
+        # Load all mixed spots
+        all_mixed_spots = self.load_all_rounds_spots_mp(table_type=table_type)
+        # coreg_spots are in ophys_mfish_match_df 'ls_id' col, need to match 'cell_id' col in all_mixed_spots
+        coreg_spots = all_mixed_spots[
+            all_mixed_spots["cell_id"].isin(ophys_mfish_match_df["ls_id"])
+        ]
+        print(f"Number of coregistered mixed spots: {len(coreg_spots)}")
+        return coreg_spots
 
     def create_cell_gene_matrix(self, unmixed=True, rounds=None):
         """
@@ -441,11 +619,15 @@ class HCRDataset:
         dataframes = {}  # Store dataframes to avoid re-reading
 
         for round_key in spot_files.keys():
-            if unmixed:
-                # Read the unmixed cell-by-gene data
-                df = pd.read_csv(spot_files[round_key].unmixed_cxg)
-            else:
-                df = pd.read_csv(spot_files[round_key].mixed_cxg)
+            try:
+                if unmixed:
+                    # Read the unmixed cell-by-gene data
+                    df = pd.read_csv(spot_files[round_key].unmixed_cxg)
+                else:
+                    df = pd.read_csv(spot_files[round_key].mixed_cxg)
+            except FileNotFoundError:
+                print(f"Warning: Spot file for round {round_key} does not exist.")
+                continue
 
             # Store the dataframe and genes for this round
             dataframes[round_key] = df
@@ -557,13 +739,15 @@ class HCRDataset:
     #             return data[data['cell_id'].isin(cell_ids)]
     #        return data
 
-    def create_channel_gene_table(self, spots_only=True):
+    def create_channel_gene_table(self, spots_only=True, label_duplicate_genes=False):
         """Create channel-gene mapping table from processing manifests."""
         processing_manifests = {
             k: round_obj.processing_manifest for k, round_obj in self.rounds.items()
         }
         return create_channel_gene_table_from_manifests(
-            processing_manifests, spots_only=spots_only
+            processing_manifests,
+            spots_only=spots_only,
+            label_duplicate_genes=label_duplicate_genes,
         )
 
     def get_segmentation_resolutions(self, round_key=None):
@@ -678,6 +862,23 @@ class HCRDataset:
                     f"    {channel}: spots {'✓' if spots_exist else '✗'}, stats files: {stats_count}"
                 )
 
+    def _print_tile_alignment_info(self):
+        """Print tile alignment file information."""
+        alignment_rounds = {
+            k: round_obj
+            for k, round_obj in self.rounds.items()
+            if round_obj.tile_alignment_files is not None
+        }
+        if not alignment_rounds:
+            return
+        print("\nTile alignment files by round:")
+        for round_key, round_obj in alignment_rounds.items():
+            xml_files = round_obj.tile_alignment_files
+            print(f"  {round_key}:")
+            for file_type, file_path in xml_files.__dict__.items():
+                if file_path:
+                    print(f"    {file_type}: {file_path.name}")
+
     def _print_file_status(self):
         """Print file existence status."""
         print("\nFile Status:")
@@ -718,11 +919,18 @@ class HCRDataset:
                     f"    Spot detection: {detection_count} of {len(round_obj.spot_detection_files)} channels exist"
                 )
 
+            if round_obj.tile_alignment_files:
+                xml_files = round_obj.tile_alignment_files
+                print(
+                    f"    Tile alignment files: {', '.join(f.name for f in xml_files.__dict__.values() if f)}"
+                )
+
     def summary(self):
         """Print a summary of the dataset."""
         self._print_basic_info()
         self._print_segmentation_info()
         self._print_spot_detection_info()
+        self._print_tile_alignment_info()
         self._print_file_status()
 
     def __dir__(self):
@@ -769,6 +977,40 @@ class HCRDataset:
 # ------------------------------------------------------------------------------------------------
 
 
+def _load_spots_for_round(round_item, table_type="mixed_spots"):
+    """
+    Helper function for multiprocessing to load spots for a single round.
+
+    Parameters:
+    -----------
+    round_item : tuple
+        Tuple of (round_key, round_obj)
+    table_type : str
+        Type of spots to load ('mixed_spots' or 'unmixed_spots')
+
+    Returns:
+    --------
+    pd.DataFrame
+        DataFrame with spots data including round column
+    """
+    round_key, round_obj = round_item
+    print(f"Loading {table_type} for round {round_key}: {round_obj.name}\n")
+
+    # Get the appropriate spot file path
+    spot_file_path = getattr(round_obj.spot_files, table_type)
+
+    # check if path exists
+    if spot_file_path is None or not spot_file_path.exists():
+        print(f"Warning: Spot file {spot_file_path} does not exist for round {round_key}")
+        return pd.DataFrame()  # Return empty DataFrame if file does not exist
+
+    with open(spot_file_path, "rb") as f:
+        spots_data = pkl.load(f)
+        spots_data["round"] = round_key
+
+    return spots_data
+
+
 def create_hcr_dataset(round_dict: dict, data_dir: Path, mouse_id: str = None):
     """
     Create a complete HCRDataset from round dictionary and data directory.
@@ -792,6 +1034,7 @@ def create_hcr_dataset(round_dict: dict, data_dir: Path, mouse_id: str = None):
     processing_manifests = get_processing_manifests(round_dict, data_dir)
     segmentation_files = get_segmentation_files(round_dict, data_dir)
     spot_detection_files = get_spot_detection_files(round_dict, data_dir)
+    tile_alignment_files = get_tile_alignment_files(round_dict, data_dir)
 
     # Create HCRRound objects
     rounds = {}
@@ -804,6 +1047,7 @@ def create_hcr_dataset(round_dict: dict, data_dir: Path, mouse_id: str = None):
             processing_manifest=processing_manifests.get(round_key, {}),
             segmentation_files=segmentation_files.get(round_key),
             spot_detection_files=spot_detection_files.get(round_key, {}),
+            tile_alignment_files=tile_alignment_files.get(round_key),
         )
 
     # Load metadata if available
@@ -923,28 +1167,28 @@ def create_channel_gene_table(spot_files: dict, spots_only=True) -> pd.DataFrame
 
             for channel, details in gene_dict.items():
                 data.append(
-                    {"Channel": channel, "Gene": details.get("gene", ""), "Round": round_key}
+                    {"channel": channel, "gene": details.get("gene", ""), "round": round_key}
                 )
 
     # sort by round then channel
-    data.sort(key=lambda x: (x["Round"], x["Channel"]))
+    data.sort(key=lambda x: (x["round"], x["channel"]))
 
     if spots_only:
         # drop Channel = 405 and Gene = Syto59
         data = [
             entry
             for entry in data
-            if not (entry["Channel"] == "405" and entry["Gene"] == "Syto59")
+            if not (entry["channel"] == "405" and entry["gene"] == "Syto59")
         ]
     # for duplicate genes, append the round name to the gene
     for entry in data:
-        if entry["Gene"] in [d["Gene"] for d in data if d["Round"] != entry["Round"]]:
-            entry["Gene"] += f"-{entry['Round']}"
+        if entry["gene"] in [d["gene"] for d in data if d["round"] != entry["round"]]:
+            entry["gene"] += f"-{entry['round']}"
     return pd.DataFrame(data)
 
 
 def create_channel_gene_table_from_manifests(
-    processing_manifests: Dict[str, dict], spots_only=True
+    processing_manifests: Dict[str, dict], spots_only=True, label_duplicate_genes=True
 ) -> pd.DataFrame:
     """
     Create a table of Channel, Gene, and Round from the "gene_dict" key in the processing manifests for each round.
@@ -981,9 +1225,10 @@ def create_channel_gene_table_from_manifests(
         ]
 
     # For duplicate genes, append the round name to the gene
-    for entry in data:
-        if entry["Gene"] in [d["Gene"] for d in data if d["Round"] != entry["Round"]]:
-            entry["Gene"] += f"-{entry['Round']}"
+    if label_duplicate_genes:
+        for entry in data:
+            if entry["Gene"] in [d["Gene"] for d in data if d["Round"] != entry["Round"]]:
+                entry["Gene"] += f"-{entry['Round']}"
 
     return pd.DataFrame(data)
 
@@ -1027,23 +1272,77 @@ def get_segmentation_files(round_dict: dict, data_dir: Path):
         # Also check for alternate name: segmentation_mask_transformed_level_2.zarr
         mask_transformed_level2_path = folder_path / "segmentation_mask_transformed_level_2.zarr"
 
-        if mask_orig_res_path.exists():
-            segmentation_masks["2"] = mask_orig_res_path
+        if mask_transformed_level2_path.exists() and mask_orig_res_path.exists():
+            print(
+                "WARNING: multiple types of segmentation masks found. Double check the data is"
+                "spots/segmentation data is correct for you. Defaulting to transformed masks."
+            )
+            segmentation_masks["2"] = mask_transformed_level2_path
         elif mask_transformed_level2_path.exists():
             segmentation_masks["2"] = mask_transformed_level2_path
+        elif mask_orig_res_path.exists():
+            segmentation_masks["2"] = mask_orig_res_path
 
         # Check for cell centroids
         centroids_path = folder_path / "cell_centroids.npy"
         if not centroids_path.exists():
             centroids_path = None
 
+        # check for metrics.pkl
+        metrics_path = folder_path / "metrics.pickle"
+        if not metrics_path.exists():
+            metrics_path = None
+
         segmentation_files[key] = SegmentationFiles(
-            segmentation_masks=segmentation_masks, cell_centroids=centroids_path
+            segmentation_masks=segmentation_masks,
+            cell_centroids=centroids_path,
+            metrics_path=metrics_path,
         )
 
     return segmentation_files
 
 
+def get_tile_alignment_files(round_dict: dict, data_dir: Path):
+    """
+    Get TileAlignmentFiles for each round based on a dictionary mapping round keys to folder names.
+    Parameters:
+    -----------
+    round_dict : dict
+        Dictionary mapping round keys (e.g., 'R1', 'R2') to folder names containing the data.
+    data_dir : Path
+        Path to the directory containing the round folders.
+    Returns:
+    --------
+    dict
+        Dictionary mapping round keys to TileAlignmentFiles objects containing paths to alignment files.
+    """
+    tile_alignment_files = {}
+    for key, folder in round_dict.items():
+        stitching_path = data_dir / folder / "image_tile_fusing" / "metadata" / "stitching"
+
+        def check_exist(path):
+            """check exist"""
+            return path if path.exists() else None
+
+        tile_alignment_files[key] = TileAlignmentFiles(
+            raw_single_xml=check_exist(
+                stitching_path / "stitching_single_channel_updated_remote.xml"
+            ),
+            raw_single_tile_subset_xml=check_exist(
+                stitching_path / "stitching_single_channel_updated_tile_subset_remote.xml"
+            ),
+            raw_spot_xml=check_exist(
+                stitching_path / "stitching_spot_channels_updated_remote.xml"
+            ),
+            raw_spot_tile_subset_xml=check_exist(
+                stitching_path / "stitching_spot_channels_updated_tile_subset_remote.xml"
+            ),
+            pc_xml=check_exist(stitching_path / "phase_correlation_stitching" / "bigstitcher.xml"),
+            ip_xml=check_exist(stitching_path / "interest_point_stitching" / "bigstitcher_0.xml"),
+        )
+    return tile_alignment_files
+
+
 def get_spot_detection_files(round_dict: dict, data_dir: Path):
     """
     Get SpotDetection objects for each round and channel based on a dictionary mapping round keys to folder names.
